"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ShopifyAdmin = void 0;
const cockatiel_1 = require("cockatiel");
const graphql_request_1 = require("graphql-request");
const debug = require("debug")("shopify-client");
const graphql_1 = require("./graphql");
const ThrottleError_1 = require("./ThrottleError");
const BadGatewayError_1 = require("./BadGatewayError");
const WebhookSubscriber_1 = require("./WebhookSubscriber");
// retry function 5 times with an exponential backoff (initially 1 second)
const retryPolicy = cockatiel_1.Policy.handleType(ThrottleError_1.ThrottleError)
    .orType(BadGatewayError_1.BadGatewayError)
    .retry()
    .attempts(5)
    .exponential({
    initialDelay: 1000,
});
class ShopifyAdmin {
    constructor(options) {
        this.activeCredentialIndex = 0;
        const shopDomain = options.shopName.includes(".myshopify.com")
            ? options.shopName
            : `${options.shopName}.myshopify.com`;
        const apiEndpoint = `https://${shopDomain}/admin/api${options.apiVersion ? `/${options.apiVersion}` : ""}/graphql.json`;
        const client = new graphql_request_1.GraphQLClient(apiEndpoint);
        this.credentials = Array.isArray(options.credentials)
            ? options.credentials
            : [options.credentials];
        if (0 === this.credentials.length) {
            throw new Error(`you must provide at least one credential to instantiate the shopify service`);
        }
        const requester = (doc, variables) => __awaiter(this, void 0, void 0, function* () {
            return yield retryPolicy.execute((context) => __awaiter(this, void 0, void 0, function* () {
                var _a, _b, _c, _d, _e;
                try {
                    return yield client.request(doc, variables, {
                        "X-Shopify-Access-Token": this.getCycledCredential().password,
                    });
                }
                catch (e) {
                    if ("THROTTLED" === ((_d = (_c = (_b = (_a = e.response) === null || _a === void 0 ? void 0 : _a.errors) === null || _b === void 0 ? void 0 : _b[0]) === null || _c === void 0 ? void 0 : _c.extensions) === null || _d === void 0 ? void 0 : _d.code)) {
                        debug(`shopify request throttled on attempt ${context.attempt}`);
                        throw new ThrottleError_1.ThrottleError();
                    }
                    // when under load, shopify sometimes returns a 502 response, rather than a throttle response
                    if (502 === ((_e = e.response) === null || _e === void 0 ? void 0 : _e.status)) {
                        debug(`shopify request returned bad gateway error on attempt ${context.attempt}`);
                        throw new BadGatewayError_1.BadGatewayError();
                    }
                    debug(`unknown shopify response exception`, e);
                    throw e;
                }
            }));
        });
        // user provided sdk
        this.sdk = options.getSdk(requester);
        // internal sdk
        this.internalSdk = graphql_1.getSdk(requester);
    }
    getCycledCredential() {
        this.activeCredentialIndex++;
        const cycledCredentialIndex = this.activeCredentialIndex % this.credentials.length;
        const cycledCredential = this.credentials[cycledCredentialIndex];
        if (!cycledCredential) {
            throw new Error(`credential with index ${cycledCredentialIndex} could not be resolved in shopify service`);
        }
        return cycledCredential;
    }
    get client() {
        return this.sdk;
    }
    get internalClient() {
        return this.internalSdk;
    }
    get webhookSubscriber() {
        return new WebhookSubscriber_1.WebhookSubscriber(this);
    }
}
exports.ShopifyAdmin = ShopifyAdmin;
